#include "pch.h"
#include "AudioFilePlayer.h"

using namespace Microsoft::WRL;
using namespace Platform;
using namespace Windows::Storage::Streams;

AudioFilePlayer::AudioFilePlayer(ComPtr<IXAudio2> pXAudio2, IXAudio2Voice* pDestinationVoice, IRandomAccessStream^ randomAccessStream)
{
    // Start up the Media Foundation
    HRESULT hresult = MFStartup(MF_VERSION);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFStartup failure");

    // Create a IMFByteStream to wrap the IRandomAccessStream
    ComPtr<IMFByteStream> mfByteStream;
    hresult = MFCreateMFByteStreamOnStreamEx((IUnknown *)randomAccessStream, &mfByteStream);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFCreateMFByteStreamOnStreamEx failure");

    // Create an attribute for low latency operation
    ComPtr<IMFAttributes> mfAttributes;
    hresult = MFCreateAttributes(&mfAttributes, 1);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFCreateAttributes failure");

    hresult = mfAttributes->SetUINT32(MF_LOW_LATENCY, TRUE);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "IMFAttributes->SetUINT32 failure");

    // Create the IMFSourceReader and save as field
    hresult = MFCreateSourceReaderFromByteStream(mfByteStream.Get(), mfAttributes.Get(), &mfSourceReader);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFCreateSourceReaderFromByteStream failure");

    // Create an IMFMediaType for setting the desired format
    ComPtr<IMFMediaType> mfMediaType;
    hresult = MFCreateMediaType(&mfMediaType);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFCreateMediaType failure");

    hresult = mfMediaType->SetGUID(MF_MT_MAJOR_TYPE, MFMediaType_Audio);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFMediaType->SetGUID failure");

    hresult = mfMediaType->SetGUID(MF_MT_SUBTYPE, MFAudioFormat_Float);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFMediaType->SetGUID failure");

    // Set the media type in the source reader
    hresult = mfSourceReader->SetCurrentMediaType(MF_SOURCE_READER_FIRST_AUDIO_STREAM, 
                                                  0, mfMediaType.Get());
    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFSourceReader->SetCurrentMediaType failure");

    // Get the Media Foundation media type
    hresult = mfSourceReader->GetCurrentMediaType(MF_SOURCE_READER_FIRST_AUDIO_STREAM, &mfMediaType);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFSourceReader->GetCurrentMediaType failure");

    // Create a WAVEFORMATEX from the media type
    WAVEFORMATEX* pWaveFormat;
    unsigned int waveFormatLength;
    hresult = MFCreateWaveFormatExFromMFMediaType(mfMediaType.Get(), &pWaveFormat, &waveFormatLength);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFCreateWaveFormatExFromMFMediaType failure");

    // Create the XAudio2 source voice, routing it to the destination voice
    XAUDIO2_SEND_DESCRIPTOR sendDescriptor;
    sendDescriptor.Flags = 0;
    sendDescriptor.pOutputVoice = pDestinationVoice;

    XAUDIO2_VOICE_SENDS sendList;
    sendList.SendCount = 1;
    sendList.pSends = &sendDescriptor;
    hresult = pXAudio2->CreateSourceVoice(&pSourceVoice, pWaveFormat,
                                          XAUDIO2_VOICE_NOPITCH, 1.0f, this, &sendList);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "pXAudio2->CreateSourceVoice failure");

    // Free the memory allocated by function
    CoTaskMemFree(pWaveFormat);

    // Submit two buffers
    SubmitBuffer();
    SubmitBuffer();

    endOfFile = false;
}

AudioFilePlayer::~AudioFilePlayer()
{
    pSourceVoice->DestroyVoice();

    HRESULT hresult = MFShutdown();

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFShutdown failure");
}

void AudioFilePlayer::Start()
{
    pSourceVoice->Start();
}

void AudioFilePlayer::Stop()
{
    pSourceVoice->Stop();
}

void AudioFilePlayer::SubmitBuffer()
{
    // Get the next block of audio data
    int audioBufferLength;
    byte * pAudioBuffer = GetNextBlock(&audioBufferLength);

    if (pAudioBuffer != nullptr)
    {
        // Create an XAUDIO2_BUFFER for submitting audio data
        XAUDIO2_BUFFER buffer = {0};
        buffer.AudioBytes = audioBufferLength;
        buffer.pAudioData = pAudioBuffer;
        buffer.pContext = pAudioBuffer;
        HRESULT hresult = pSourceVoice->SubmitSourceBuffer(&buffer);

        if (FAILED(hresult))
            throw ref new COMException(hresult, "IXAudio2SourceVoice->SubmitSourceBuffer failure");
    }
}

byte * AudioFilePlayer::GetNextBlock(int * pAudioBufferLength)
{
    // Get an IMFSample object
    ComPtr<IMFSample> mfSample;
    DWORD flags = 0;
    HRESULT hresult = mfSourceReader->ReadSample(MF_SOURCE_READER_FIRST_AUDIO_STREAM, 
                                                 0, nullptr, &flags, nullptr, &mfSample);
    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFSourceReader->ReadSample failure");

    // Check if we're at the end of the file
    if (flags & MF_SOURCE_READERF_ENDOFSTREAM)
    {
        endOfFile = true;
        *pAudioBufferLength = 0;
        return nullptr;
    }

    // If not, convert the data to a contiguous buffer
    ComPtr<IMFMediaBuffer> mfMediaBuffer;
    hresult = mfSample->ConvertToContiguousBuffer(&mfMediaBuffer);

    if (FAILED(hresult))
         throw ref new COMException(hresult, "IMFSample->ConvertToContiguousBuffer failure");

    // Lock the audio buffer and copy the samples to local memory
    uint8 * pAudioData = nullptr;
    DWORD audioDataLength = 0;
    hresult = mfMediaBuffer->Lock(&pAudioData, nullptr, &audioDataLength); 

    if (FAILED(hresult))
        throw ref new COMException(hresult, "IMFMediaBuffer->Lock failure");

    byte * pAudioBuffer = new byte[audioDataLength];
    CopyMemory(pAudioBuffer, pAudioData, audioDataLength);
    hresult = mfMediaBuffer->Unlock();

    if (FAILED(hresult))
        throw ref new COMException(hresult, "IMFMediaBuffer->Unlock failure");

    *pAudioBufferLength = audioDataLength;
    return pAudioBuffer;
}

// Callback method from IXAudio2VoiceCallback
void _stdcall AudioFilePlayer::OnBufferEnd(void* pContext)
{
    // Remember to free the audio buffer!
    delete[] pContext;

    // Submit a new buffer
    if (!endOfFile)
    {
        SubmitBuffer();
    }
}
