#include "pch.h"
#include "OneThirdOctaveEqualizerEffect.h"
#include "OneThirdOctaveEqualizerParameters.h"

const XAPO_REGISTRATION_PROPERTIES OneThirdOctaveEqualizerEffect::RegistrationProps = 
{
    __uuidof(OneThirdOctaveEqualizerEffect),
    L"One-Third Octave Equalizer Effect",
    L"Coded by Charles Petzold",
    1,      // major version number
    0,      // minor version number
    XAPOBASE_DEFAULT_FLAG | XAPO_FLAG_INPLACE_REQUIRED,
    1,      // min input buffer count
    1,      // max input buffer count
    1,      // min output buffer count
    1       // max output buffer count
};

OneThirdOctaveEqualizerEffect * OneThirdOctaveEqualizerEffect::Create()
{
    // Create and initialize three effect parameters
    OneThirdOctaveEqualizerParameters * pParameterBlocks = 
                            new OneThirdOctaveEqualizerParameters[3];

    for (int i = 0; i < 3; i++)
        for (int band = 0; band < 26; band++)
            pParameterBlocks[i].Amplitude[band] = 1.0f;
    
    // Create the effect
    return new OneThirdOctaveEqualizerEffect(
                            &RegistrationProps, 
                            (byte *) pParameterBlocks, 
                            sizeof(OneThirdOctaveEqualizerParameters), 
                            false);
}

OneThirdOctaveEqualizerEffect::OneThirdOctaveEqualizerEffect(const XAPO_REGISTRATION_PROPERTIES * pRegProperties, 
                                                             BYTE * pParameterBlocks,
                                                             UINT32 uParameterBlockByteSize,
                                                             BOOL fProducer) 
                                                             :
                                                             CXAPOParametersBase(pRegProperties, 
                                                                                 pParameterBlocks, 
                                                                                 uParameterBlockByteSize, 
                                                                                 fProducer),
                                                             pxp(nullptr),
                                                             pxpp(nullptr),
                                                             pyp(nullptr),
                                                             pypp(nullptr)
{
    // Save parameters blocks for deletion in destructor
    this->pParameterBlocks = pParameterBlocks;
}

OneThirdOctaveEqualizerEffect::~OneThirdOctaveEqualizerEffect()
{
    FreeArrays();

    if (pParameterBlocks != nullptr)
        delete[] pParameterBlocks;
}

void OneThirdOctaveEqualizerEffect::FreeArrays()
{
    if (pxp != nullptr)
        delete[] pxp;

    if (pxpp != nullptr)
        delete[] pxpp;

    if (pyp != nullptr)
        delete[] pyp;

    if (pypp != nullptr)
        delete[] pypp;
}

HRESULT OneThirdOctaveEqualizerEffect::LockForProcess(UINT32 inpParamCount,
                                                      const XAPO_LOCKFORPROCESS_BUFFER_PARAMETERS  *pInpParams,
                                                      UINT32 outParamCount,
                                                      const XAPO_LOCKFORPROCESS_BUFFER_PARAMETERS  *pOutParams)
{
    waveFormat = * pInpParams[0].pFormat;

    // Verify sufficiently high sample rate
    if (waveFormat.nSamplesPerSec < 44100)
        return E_NOTIMPL;

    // Calculate the filter factors
    Q = 4.318f;       // 1/3 octave

    static float frequencies[26] = 
    {
        20.0f, 25.0f, 31.5f, 40.0f, 50.0f, 63.0f, 80.0f, 100.0f, 125.0f,
        160.0f, 200.0f, 250.0f, 320.0f, 400.0f, 500.0f, 630.0f, 800.0f, 1000.0f,
        1250.0f, 1600.0f, 2000.0f, 2500.0f, 3150.0f, 4000.0f, 5000.0f, 6300.0f
    };

    for (int band = 0; band < 26; band++)
    {
        float frequency = frequencies[band];
        float omega = 2 * 3.14159f * frequency / waveFormat.nSamplesPerSec;
        float alpha = sin(omega) / (2 * Q);

        a0[band] = 1 + alpha;
        a1[band] = -2 * cos(omega);
        a2[band] = 1 - alpha;
        b0[band] = Q * alpha;       // == sin(omega) / 2;
        b1[band] = 0;
        b2[band] = -Q * alpha;      // == -sin(omega) / 2;
    }

    FreeArrays();

    // Allocate memory for saving previous samples
    int channels = waveFormat.nChannels;

    pxp = new float[channels];
    pxpp = new float[channels];
    pyp = new float[channels * 26];
    pypp = new float[channels * 26];

    for (int channel = 0; channel < channels; channel++)
    {
        pxp[channel] = 0.0f;
        pxpp[channel] = 0.0f;
        
        for (int band = 0; band < 26; band++)
        {
            pyp[channels * band + channel] = 0.0f;
            pypp[channels * band + channel] = 0.0f;
        }
    }

    return CXAPOBase::LockForProcess(inpParamCount, pInpParams, outParamCount, pOutParams);
}

void OneThirdOctaveEqualizerEffect::Process(UINT32 inpParamCount,
                                            const XAPO_PROCESS_BUFFER_PARAMETERS *pInpParam,
                                            UINT32 outParamCount,
                                            XAPO_PROCESS_BUFFER_PARAMETERS *pOutParam,
                                            BOOL isEnabled)
{
    // Get effect parameters
    OneThirdOctaveEqualizerParameters * pEqualizerParams = 
        (OneThirdOctaveEqualizerParameters *)CXAPOParametersBase::BeginProcess();

    // Get buffer pointers and other information
    const float * pSrc = static_cast<float *>(pInpParam[0].pBuffer);
    float * pDst = static_cast<float *>(pOutParam[0].pBuffer);
    int frameCount = pInpParam[0].ValidFrameCount;
    int numChannels = waveFormat.nChannels;

    switch(pInpParam[0].BufferFlags)
    {
    case XAPO_BUFFER_VALID:
        for (int frame = 0; frame < frameCount; frame++)
        {
            for (int channel = 0; channel < numChannels; channel++)
            {
                int index = numChannels * frame + channel;

                // Do very little if filter is disabled
                if (!isEnabled)
                {
                    pDst[index] = pSrc[index];
                    continue;
                }

                // Get previous inputs
                float x = pSrc[index];
                float xp = pxp[channel];
                float xpp = pxpp[channel];

                // Initialize accumulated value
                float accum = 0;

                for (int band = 0; band < 26; band++)
                {
                    int bandIndex = numChannels * band + channel;

                    // Get previous outputs
                    float yp = pyp[bandIndex];
                    float ypp = pypp[bandIndex];

                    // Calculate filter output
                    float y = (b0[band] * x + b1[band] * xp + b2[band] * xpp
                                            - a1[band] * yp - a2[band] * ypp) / a0[band];

                    // Accumulate amplitude-adjusted filter output
                    accum += y * pEqualizerParams->Amplitude[band];

                    // Save previous output values
                    pypp[bandIndex] = yp;
                    pyp[bandIndex] = y;
                }

                // Save previous input values
                pxpp[channel] = xp;
                pxp[channel] = x;

                // Save final value adjusted for filter gain
                pDst[index] = accum / Q;
            }
        }
        break;

    case XAPO_BUFFER_SILENT:
        break;
    }

    // Set output parameters
    pOutParam[0].ValidFrameCount = pInpParam[0].ValidFrameCount;
    pOutParam[0].BufferFlags = pInpParam[0].BufferFlags;

    CXAPOParametersBase::EndProcess();
}





