#include "pch.h"
#include "OneSecondEchoEffect.h"

const XAPO_REGISTRATION_PROPERTIES OneSecondEchoEffect::RegistrationProps = 
{
    __uuidof(OneSecondEchoEffect),
    L"One-Second Echo Effect",
    L"Coded by Charles Petzold",
    1,      // major version number
    0,      // minor version number
    XAPOBASE_DEFAULT_FLAG | XAPO_FLAG_INPLACE_REQUIRED,
    1,      // min input buffer count
    1,      // max input buffer count
    1,      // min output buffer count
    1       // max output buffer count
};

HRESULT OneSecondEchoEffect::LockForProcess(UINT32 inpParamCount,
                                            const XAPO_LOCKFORPROCESS_BUFFER_PARAMETERS  *pInpParams,
                                            UINT32 outParamCount,
                                            const XAPO_LOCKFORPROCESS_BUFFER_PARAMETERS  *pOutParams)
{
    waveFormat = * pInpParams[0].pFormat;

    // Size delay buffer appropriate for the WAVEFORMAT
    delayLength = waveFormat.nSamplesPerSec;    
    int numDelaySamples = waveFormat.nChannels * waveFormat.nSamplesPerSec;
    delayBuffer.resize(numDelaySamples);

    // Initialize to all zeros
    for (int sample = 0; sample < numDelaySamples; sample++)
        delayBuffer[sample] = 0.0f;

    delayIndex = 0;

    return CXAPOBase::LockForProcess(inpParamCount, pInpParams, outParamCount, pOutParams);
}

void OneSecondEchoEffect::Process(UINT32 inpParamCount,
                                  const XAPO_PROCESS_BUFFER_PARAMETERS *pInpParams,
                                  UINT32 outParamCount,
                                  XAPO_PROCESS_BUFFER_PARAMETERS *pOutParams,
                                  BOOL isEnabled)
{
    const float * pSrc = static_cast<float *>(pInpParams[0].pBuffer);
    float * pDst = static_cast<float *>(pOutParams[0].pBuffer);
    int frameCount = pInpParams[0].ValidFrameCount;
    int numChannels = waveFormat.nChannels;
    bool isSourceValid = pInpParams[0].BufferFlags == XAPO_BUFFER_VALID;

    for (int frame = 0; frame < frameCount; frame++)
    {
        for (int channel = 0; channel < numChannels; channel++)
        {
            // Get sample based on XAPO_BUFFER_VALID flag
            int index = numChannels * frame + channel;
            float source = isSourceValid ? pSrc[index] : 0.0f;

            // Combine sample with contents of delay buffer and save back
            int delayBufferIndex = numChannels * delayIndex + channel;
            float echo = 0.5f * source + 0.5f * delayBuffer[delayBufferIndex];
            delayBuffer[delayBufferIndex] = echo;

            // Transfer to destination buffer
            pDst[index] = isEnabled ? echo : source;
        }
        delayIndex = (delayIndex + 1) % delayLength;
    }

    pOutParams[0].BufferFlags = XAPO_BUFFER_VALID;
    pOutParams[0].ValidFrameCount = pInpParams[0].ValidFrameCount;
}
